import {
  KeyDataPair,
  MessageData,
  DocumentData,
  SealHeadData,
  SealIndexObject,
} from "../types";
import { vereign } from "../generated/qrcode_data_pb";

import {
  arrayBufferToBase64,
  base64ToArrayBuffer,
  ensureUint8Array,
} from "../utils/common";

import { CryptoService } from "../index";

const DocumentDataMessageV1 = vereign.protobuf.qrcode_data.DocumentData_V1;
const EmailDataMessageV1 = vereign.protobuf.qrcode_data.EmailData_V1;
const KeyDataMessageV1 = vereign.protobuf.qrcode_data.KeyData_V1;
const WrapperDataMessage = vereign.protobuf.qrcode_data.WrapperData;
const SealIndexObjectMessageV1 =
  vereign.protobuf.qrcode_data.SealIndexObject_V1;
const SealHeadMessageV1 = vereign.protobuf.qrcode_data.SealDocumentHeadData_V1;

const EMAIL_DATA_V1 = "EmailData_V1";
const DOCUMENT_DATA_V1 = "DocumentData_V1";

interface WrappedData {
  version: number;
  data: ArrayBuffer;
  className: string;
}

const encodeEmailData = (emailData: MessageData): string => {
  const emailDataMessage = EmailDataMessageV1.fromObject(emailData);
  const emailDataBuffer = EmailDataMessageV1.encode(emailDataMessage).finish();

  const wrappedData = {
    version: 1,
    className: EMAIL_DATA_V1,
    data: emailDataBuffer,
  };

  const wrappedDataMessage = WrapperDataMessage.fromObject(wrappedData);
  const wrappedDataBuffer =
    WrapperDataMessage.encode(wrappedDataMessage).finish();
  return arrayBufferToBase64(wrappedDataBuffer);
};

const decodeEmailData = (
  binary: string | Uint8Array | ArrayBuffer
): MessageData => {
  const wrappedDataMessage = WrapperDataMessage.decode(
    ensureUint8Array(binary)
  );
  const wrappedData = WrapperDataMessage.toObject(
    wrappedDataMessage
  ) as WrappedData;

  if (wrappedData.className === EMAIL_DATA_V1) {
    const emailDataMessage = EmailDataMessageV1.decode(
      ensureUint8Array(wrappedData.data)
    );

    const messageData = EmailDataMessageV1.toObject(emailDataMessage, {
      bytes: String,
      arrays: true, // populates empty arrays (repeated fields) even if defaults=false
      objects: true, // populates empty objects (map fields) even if defaults=false
    }) as MessageData;

    return messageData;
  } else {
    return null;
  }
};

const encodeKeyDataPair = (keyData: KeyDataPair): string => {
  const keyDataMessage = KeyDataMessageV1.fromObject(keyData);
  const buffer = KeyDataMessageV1.encode(keyDataMessage).finish();

  return arrayBufferToBase64(buffer);
};

const decodeKeyDataPair = (
  binary: string | Uint8Array | ArrayBuffer
): KeyDataPair => {
  const keyDataPairMessage = KeyDataMessageV1.decode(ensureUint8Array(binary));
  return KeyDataMessageV1.toObject(keyDataPairMessage, {
    bytes: String,
  }) as KeyDataPair;
};

const breakQrCodeData = (
  data: ArrayBuffer,
  headBytesSize = 32
): { head: ArrayBuffer; tail: ArrayBuffer } => {
  if (data.byteLength <= headBytesSize) {
    return {
      head: data,
      tail: null,
    };
  } else {
    const dataUint8Array = new Uint8Array(data);
    return {
      head: dataUint8Array.slice(0, headBytesSize).buffer,
      tail: dataUint8Array.slice(headBytesSize, dataUint8Array.length).buffer,
    };
  }
};

const assembleQrCodeData = (
  head: ArrayBuffer,
  tail: ArrayBuffer
): ArrayBuffer => {
  return Buffer.concat([Buffer.from(head), Buffer.from(tail)]);
};

const calculateQRCodeSignature = async (
  accountPrivateKey: string,
  qrCodeHash: string
): Promise<string> => {
  const signature = await CryptoService.signRSA(
    accountPrivateKey,
    Buffer.from(qrCodeHash)
  );

  return arrayBufferToBase64(signature);
};

const verifyQrCodeSignature = async (
  publicKey: string,
  qrCodeSignature: string,
  recipientQRCodeHash: string
): Promise<boolean> => {
  const qrCodeSignatureAsBuffer = base64ToArrayBuffer(qrCodeSignature);
  const recipientQRCodeHashAsBuffer = Buffer.from(recipientQRCodeHash);
  const result = await CryptoService.verifyRSASignature(
    publicKey,
    recipientQRCodeHashAsBuffer,
    qrCodeSignatureAsBuffer
  );

  return result;
};

const decodeDocumentData = (
  binary: string | Uint8Array | ArrayBuffer
): DocumentData => {
  const wrappedDataMessage = WrapperDataMessage.decode(
    ensureUint8Array(binary)
  );
  const wrappedData = WrapperDataMessage.toObject(
    wrappedDataMessage
  ) as WrappedData;

  if (wrappedData.className === DOCUMENT_DATA_V1) {
    const documentDataMessage = DocumentDataMessageV1.decode(
      ensureUint8Array(wrappedData.data)
    );

    const messageData = DocumentDataMessageV1.toObject(documentDataMessage, {
      bytes: String,
    }) as DocumentData;

    return messageData;
  }

  return null;
};

const encodeDocumentData = (documentData: DocumentData): string => {
  const documentDataMessage = DocumentDataMessageV1.fromObject(documentData);
  const documentDataBuffer =
    DocumentDataMessageV1.encode(documentDataMessage).finish();

  const wrappedData = {
    version: 1,
    className: DOCUMENT_DATA_V1,
    data: documentDataBuffer,
  };

  const wrappedDataMessage = WrapperDataMessage.fromObject(wrappedData);
  const wrappedDataBuffer =
    WrapperDataMessage.encode(wrappedDataMessage).finish();
  return arrayBufferToBase64(wrappedDataBuffer);
};

const encodeSealHead = (sealHeadData: SealHeadData): string => {
  const sealHeadDataMessage = SealHeadMessageV1.fromObject(sealHeadData);
  const buffer = SealHeadMessageV1.encode(sealHeadDataMessage).finish();

  return arrayBufferToBase64(buffer);
};

const decodeSealHead = (
  binary: string | Uint8Array | ArrayBuffer
): SealHeadData => {
  const sealHeadDataMessage = SealHeadMessageV1.decode(
    ensureUint8Array(binary)
  );

  return SealHeadMessageV1.toObject(sealHeadDataMessage, {
    bytes: String,
  }) as SealHeadData;
};

const encodeSealIndexObject = (indexObjectData: SealIndexObject): string => {
  const sealIndexDataMessage =
    SealIndexObjectMessageV1.fromObject(indexObjectData);
  const buffer = SealIndexObjectMessageV1.encode(sealIndexDataMessage).finish();

  return arrayBufferToBase64(buffer);
};

const decodeSealIndexObject = (
  binary: string | Uint8Array | ArrayBuffer
): SealIndexObject => {
  const sealIndexDataMessage = SealIndexObjectMessageV1.decode(
    ensureUint8Array(binary)
  );

  return SealIndexObjectMessageV1.toObject(sealIndexDataMessage, {
    bytes: String,
  }) as SealIndexObject;
};

export default {
  encodeSealHead,
  decodeSealHead,
  encodeSealIndexObject,
  decodeSealIndexObject,
  encodeDocumentData,
  decodeDocumentData,
  encodeEmailData,
  decodeEmailData,
  encodeKeyDataPair,
  decodeKeyDataPair,
  breakQrCodeData,
  assembleQrCodeData,
  calculateQRCodeSignature,
  verifyQrCodeSignature,
};
