import {
  arrayBufferToBase64,
  base64ToArrayBuffer,
  CryptoService,
  DecodedCallData,
  getMerkleTreeRootHash,
} from "../../index";

const EventEmitter = require("eventemitter2");
import AeternityService from "../AeternityService/AeternityService";
import CloudflareService from "../CloudflareService";
import {
  BlockData,
  StatusesBatchData,
  BatchVerificationDetails,
  StatusData,
  TxData,
  VerificationData,
} from "../../types";
import VerificationError from "./VerificationError";
import { AxiosError } from "axios";

export const TRANSACTION_RETRIEVED = "TRANSACTION_RETRIEVED";
export const STATUS_BATCH_VERIFIED = "STATUS_BATCH_VERIFIED";
export const BLOCK_DATA_RETRIEVED = "BLOCK_DATA_RETRIEVED";

/**
 * Legacy verification method. Support for older statuses
 */
const VERIFICATION_METHOD_MERKLE_TREE = "VERIFICATION_METHOD_MERKLE_TREE";
const VERIFICATION_METHOD_SHA256_BATCH = "VERIFICATION_METHOD_SHA256_BATCH";

const STATUSES_BATCH_NAME_PREFIX = "batch";
const STATUSES_MERKLE_TREE_NAME_PREFIX = "bmt";

class VerificationService extends EventEmitter {
  private _aeternityService: AeternityService = null;
  private _cloudflareService: CloudflareService = null;

  constructor(
    cdnUrl?: string,
    cdnBucket?: string,
    aeternityNodeUrls?: string[] | string,
    aeternityCompilerUrls?: string[] | string,
    aeternityContractBytecode?: string
  ) {
    super();

    this._aeternityService = new AeternityService(
      aeternityNodeUrls,
      aeternityCompilerUrls,
      aeternityContractBytecode
    );

    this._cloudflareService = new CloudflareService(cdnUrl, cdnBucket);
  }

  async getTransaction(hash: string): Promise<TxData> {
    let tx;
    try {
      tx = await this._aeternityService.getTxDataByHash(hash);
    } catch (e) {
      this.catchVerificationError(e, `Error getting transaction`);
      return null;
    }

    return tx;
  }

  async getBlockData(blockHeight: number): Promise<BlockData> {
    let blockchainHeight;
    let blockInfo;
    try {
      if (blockHeight > -1) {
        const result = await Promise.all([
          this._aeternityService.getBlockchainHeight(),
          this._aeternityService.getBlock(blockHeight),
        ]);
        blockchainHeight = result[0];
        blockInfo = result[1];
      } else {
        // Transactions with blockHeight = -1 considered as the ones with zero confirmations
        // and pending to be settled in the block at latest height
        // Original issue: https://github.com/aeternity/aeternity/pull/3550

        blockchainHeight = await this._aeternityService.getBlockchainHeight();
        blockInfo = await this._aeternityService.getBlock(blockchainHeight);
      }
    } catch (e) {
      this.catchVerificationError(e, `Error retrieving block`);
      return null;
    }

    return {
      blockHeight: blockInfo.height,
      timestamp: blockInfo.time,
      confirmationsAmount: blockchainHeight - blockInfo.height,
    };
  }

  async verifyStatusData(statusData: StatusData): Promise<VerificationData> {
    let verificationError: VerificationError;
    let blockData: BlockData;
    let batchVerificationDetails: BatchVerificationDetails;
    let txData: TxData;

    try {
      txData = await this.getTransaction(statusData.transactionHash);
      this.emit(TRANSACTION_RETRIEVED, txData);

      const statusBatchData = await this.getStatusBatchDataForTx(txData);
      const {
        verificationMethod,
        items: batchItems,
        hash: batchHash,
      } = statusBatchData;

      let statusVerified;
      let statusPosition = -1;
      if (verificationMethod === VERIFICATION_METHOD_MERKLE_TREE) {
        ({
          verified: statusVerified,
          statusPosition,
        } = await this.verifyStatusesMerkleTree(
          statusData.statusRaw,
          statusBatchData
        ));
      } else if (verificationMethod === VERIFICATION_METHOD_SHA256_BATCH) {
        ({
          verified: statusVerified,
          statusPosition,
        } = await this.verifyStatusSHA256Batch(
          statusData.statusRaw,
          statusBatchData
        ));
      }

      batchVerificationDetails = {
        verified: statusVerified,
        batchSize: batchItems.length,
        statusPosition,
        batchHash,
      };

      this.emit(STATUS_BATCH_VERIFIED, batchVerificationDetails);

      const blockRequests = await Promise.all([
        this.getBlockData(txData.blockHeight),
        this._aeternityService.getMicroBlock(txData.blockHash),
      ]);

      blockData = blockRequests[0];
      const microBlockData = blockRequests[1];
      txData.time = microBlockData.time;

      this.emit(BLOCK_DATA_RETRIEVED, blockData);
    } catch (e) {
      if (e instanceof VerificationError) {
        verificationError = e;
      } else {
        throw e;
      }
    }

    return {
      statusData,
      batchVerificationDetails,
      blockData,
      verificationError,
      txData,
    };
  }

  async getStatusBatchDataForTx(txData: TxData): Promise<StatusesBatchData> {
    let verificationMethod;
    let batchItems;

    let decodedCallData: DecodedCallData;
    try {
      decodedCallData = await this._aeternityService.decodeContractCallData(
        txData.tx?.callData
      );
    } catch (e) {
      this.catchVerificationError(e, `Error decoding contract call data`);
    }

    const { key: batchFilename, value } = decodedCallData;
    const batchHash = value;

    const lowerCaseFilename = batchFilename.toLowerCase();
    if (lowerCaseFilename.startsWith(STATUSES_MERKLE_TREE_NAME_PREFIX)) {
      verificationMethod = VERIFICATION_METHOD_MERKLE_TREE;
    } else if (lowerCaseFilename.startsWith(STATUSES_BATCH_NAME_PREFIX)) {
      verificationMethod = VERIFICATION_METHOD_SHA256_BATCH;
    } else {
      // Unrecognized verification method
      throw new VerificationError(
        `No corresponding verification method found for batch "${batchFilename}"`
      );
    }

    try {
      if (verificationMethod === VERIFICATION_METHOD_MERKLE_TREE) {
        batchItems = await this._cloudflareService.getMerkleTree(batchFilename);
      } else if (verificationMethod === VERIFICATION_METHOD_SHA256_BATCH) {
        ({ data: batchItems } = await this._cloudflareService.fetchFile<
          string[]
        >(batchFilename));
      }
    } catch (e) {
      this.catchVerificationError(
        e,
        `Error obtaining batch of statuses for verification.`
      );
    }

    return {
      verificationMethod,
      hash: batchHash,
      items: batchItems,
    };
  }

  async verifyStatusSHA256Batch(
    status: string,
    statusesBatchData: StatusesBatchData
  ): Promise<{ verified: boolean; statusPosition: number }> {
    const statusSha256 = arrayBufferToBase64(
      await CryptoService.SHA256(status, "base64")
    );
    let verified = false;
    const statusPosition = statusesBatchData.items.indexOf(statusSha256) + 1;
    if (statusPosition > 0) {
      const bytesToHash: Uint8Array[] = [];
      let hashesTotalByteLength = 0;
      for (const statusHash of statusesBatchData.items) {
        const statusHashBytes = base64ToArrayBuffer(statusHash);
        hashesTotalByteLength += statusHashBytes.byteLength;
        bytesToHash.push(new Uint8Array(statusHashBytes));
      }
      const batchBytes = new Uint8Array(hashesTotalByteLength);
      let bytesOffset = 0;
      bytesToHash.forEach((statusHashBytes) => {
        batchBytes.set(statusHashBytes, bytesOffset);
        bytesOffset += statusHashBytes.byteLength;
      });
      const batchHash = arrayBufferToBase64(
        await CryptoService.SHA256(batchBytes)
      );

      verified = batchHash === statusesBatchData.hash;
    } else {
      throw new VerificationError(`Status not present in a batch.`);
    }

    if (!verified) {
      throw new VerificationError(`Batch of statuses not verified.`);
    }

    return {
      verified,
      statusPosition,
    };
  }

  async verifyStatusesMerkleTree(
    nodeToVerify: string,
    merkleTree: StatusesBatchData
  ): Promise<{ verified: boolean; statusPosition: number }> {
    const leaves = await Promise.all(
      merkleTree.items.map(async (x) =>
        Buffer.from(await CryptoService.SHA256(x))
      )
    );

    const rootHash = (await getMerkleTreeRootHash(leaves)).toString("base64");
    const verified = rootHash === merkleTree.hash;

    if (!verified) {
      throw new VerificationError(`Merkle tree not verified.`);
    }

    const statusPosition = merkleTree.items.indexOf(nodeToVerify) + 1;

    if (statusPosition === 0) {
      throw new VerificationError(`Status not present in merkle tree.`);
    }

    return { verified, statusPosition };
  }

  catchVerificationError(
    error: AxiosError,
    verificationFailureMessage: string
  ): void {
    if (error.response) {
      throw new VerificationError(verificationFailureMessage);
    } else {
      throw error;
    }
  }
}

VerificationService.TRANSACTION_RETRIEVED = TRANSACTION_RETRIEVED;
VerificationService.STATUS_BATCH_VERIFIED = STATUS_BATCH_VERIFIED;
VerificationService.BLOCK_DATA_RETRIEVED = BLOCK_DATA_RETRIEVED;

export default VerificationService;
