import axios, { Method } from "axios";

export interface RequestConfig {
  url?: string;
  method?: string;
  headers?: unknown;
  params?: unknown;
  data?: unknown;
}

export interface Response<T = unknown> {
  data: T;
  status: number;
  statusText: string;
  headers: unknown;
  config: RequestConfig;
}

export type Request = (config: RequestConfig) => Promise<Response>;

const defaultRequest: Request = async (config) => {
  const { url, method, headers, params, data } = config;
  try {
    const response = await axios({
      url,
      method: method as Method,
      headers,
      params,
      data,
    });

    return {
      data: response.data,
      status: response.status,
      statusText: response.statusText,
      headers: response.headers,
      config: config,
    };
  } catch (e) {
    let response: Response;

    if (e.response) {
      response = {
        data: e.response.data,
        status: e.response.status,
        statusText: e.response.statusText,
        headers: e.response.headers,
        config: config,
      };
    }

    throw new RequestError(e.message, config, response);
  }
};

export class RequestError extends Error {
  public config: RequestConfig;
  public response: Response;
  public isRequestError = true;
  constructor(message: string, config: RequestConfig, response?: Response) {
    super(message);
    this.response = response;
    this.config = config;
  }
}

export default defaultRequest;
